@@ -1,7 +1,7 @@
 import sqlalchemy as sa
 from sqlalchemy_utils.functions import create_database
-from pyg_base import cache, cfg_read, as_list, dictable, lower, last, loop, replace, Dict, is_dict, is_dictable, is_strs, is_str, is_int, is_date, dt2str, ulist, try_back, unique, is_primitive
-from pyg_encoders import as_reader, as_writer, dumps, loads
+from pyg_base import cache, cfg_read, as_list, dictable, lower, loop, replace, Dict, is_dict, is_dictable, is_strs, is_str, is_int, is_date, dt2str, ulist, try_back, unique, is_primitive
+from pyg_encoders import as_reader, as_writer, dumps, loads, encode
 from sqlalchemy import Table, Column, Integer, String, MetaData, Identity, Float, DATE, DATETIME, TIME, select, func, not_, desc, asc
 from sqlalchemy.orm import Session
 from sqlalchemy.types import NUMERIC, FLOAT, INT
@@ -10,7 +10,7 @@ from copy import copy
 from pyg_base import logger, is_regex
 from functools import partial
 import pandas as pd
-import pickle
+
 
 _id = '_id'
 _doc = 'doc'
@@ -30,6 +30,28 @@ _types = {str: String, 'str' : String,
 
 _orders = {1 : asc, True: asc, 'asc': asc, asc : asc, -1: desc, False: desc, 'desc': desc, desc: desc}
 
+def _data_and_columns(executed):
+    columns = list(executed.keys())
+    data = list(executed)
+    return data, columns
+
+def valid_session(session):
+    """
+    TO DO: How do we check if db connection expired??
+
+    Parameters
+    ----------
+    session : db session
+        sql database session
+
+    Returns
+    -------
+    bool
+        is the session still valid?
+
+    """
+    return session is not None
+
 def _relabel(res, selection, strict = True):
     """
     selection defines the CASE we want the columns to be in
@@ -52,7 +74,7 @@ def _relabel(res, selection, strict = True):
     Example
     -------
     >>> res = dict(a = 1, b = 2, c = 3)
-    >>> selection  = ['A', 'B']
+    >>> selection = ['A', 'B']
     >>> assert _relabel(res, selection) == res                  ## no replacement, strict
     >>> assert _relabel(res, selection, strict = False) == dict(A = 1, B = 2, c = 3) ## replaceme what is possible
     >>> assert _relabel(res, ['A', 'B', 'C']) == dict(A = 1, B = 2, C = 3)
@@ -116,18 +138,21 @@ def _pw_filter(col, v):
         return sa.or_(*[_pw_filter(col, i) for i in v])
     else:
         return col == v
-    
+
+@loop(list)
+def _like_within_doc(v, k):
+    return '%' + dumps(encode({k : v}))[1:-1] + '%'
 
 
-def pickle_loads(value):
-    if isinstance(value, dict):
-        return type(value)({k: pickle_loads(v) for k, v in value.items()})
-    elif isinstance(value, (list, tuple)):
-        return type(value)([pickle_loads(v) for v in value])        
-    elif isinstance(value, bytes):
-        return pickle.loads(value)
-    else:
-        return value
+# def pickle_loads(value):
+#     if isinstance(value, dict):
+#         return type(value)({k: pickle_loads(v) for k, v in value.items()})
+#     elif isinstance(value, (list, tuple)):
+#         return type(value)([pickle_loads(v) for v in value])        
+#     elif isinstance(value, bytes):
+#         return pickle.loads(value)
+#     else:
+#         return value
 
 
 @cache
@@ -137,6 +162,7 @@ def _servers():
         res[None] = res.get('null')
         res[True] = res.get('true')
         res[False] = res.get('false')
+        res['None'] = res.get('None', res[None])
     return res
 
 @cache
@@ -292,8 +318,9 @@ def _get_engine(*pairs, **connection):
 def get_engine(*pairs, **connection):
     """
     returns a sqlalchemy engine object
-    accepts either *pairs: 'driver={ODBC Driver 17 for SQL Server}'
-    or keyword arguments that look like driver = 'ODBC Driver 17 for SQL Server'    
+    accepts either 
+    *pairs: 'driver={ODBC Driver 17 for SQL Server}' or
+    **connection: keyword arguments that look like driver = 'ODBC Driver 17 for SQL Server'    
     """
     return _get_engine(*pairs, **connection)
     
@@ -310,7 +337,7 @@ def _get_table(table_name, schema, db, server, create):
 
 def sql_table(table, db = None, non_null = None, nullable = None, _id = None, schema = None, 
               server = None, reader = None, writer = None, pk = None, doc = None, mode = None, 
-              spec = None, selection = None, order = None, joint = None, create = None):
+              spec = None, selection = None, order = None, defaults = None, joint = None, create = None):
     """
     Creates a sql table. Can also be used to simply read table from the db
 
@@ -360,6 +387,8 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
                 create = False
             else:
                 create = 's'
+    defaults:
+        this is default variable that is added to the document if it does not have its keys
 
     Returns
     -------
@@ -489,7 +518,7 @@ def sql_table(table, db = None, non_null = None, nullable = None, _id = None, sc
         else:
             raise ValueError(f'table {table_name} does not exist. You need to explicitly set create=True or create="t/s/d" to mandate table creation')
     res = sql_cursor(table = tbl, schema = schema, db = db, server = server, engine = e, 
-                     reader = reader, writer = writer, 
+                     reader = reader, writer = writer, defaults = defaults,
                      pk = list(pk) if isinstance(pk, dict) else pk, doc = doc,
                      spec = spec, selection = selection, order = order, joint = joint)
     return res
@@ -797,9 +826,9 @@ class sql_cursor(object):
     >>> assert list(read_from_db.salary.values) == [100, 200, 300]
     >>> assert list(read_from_file.values) == [100, 200, 300]
     """
-    def __init__(self, table, schema = None, db = None, engine = None, server = None, 
+    def __init__(self, table, schema = None, db = None, engine = None, server = None, session = None,
                  spec = None, selection = None, order = None, joint = None, reader = None, writer = None, 
-                 pk = None, doc = None, **_):
+                 pk = None, defaults = None, doc = None, **_):
         """
         Parameters
         ----------
@@ -813,6 +842,8 @@ class sql_cursor(object):
             The sqlalchemy engine
         server : str , optional
             The server for the engine. If none, uses the default in pyg config file
+        session: db session, optional
+            An uncommitted session
         spec : sa.Expression, optional
             The "where" statement
         selection : str/list of str, optional
@@ -834,6 +865,7 @@ class sql_cursor(object):
             db = table.db if db is None else db
             engine = table.engine if engine is None else engine
             server = table.server if server is None else server
+            session = table.session if session is None else session
             spec = table.spec if spec is None else spec
             selection = table.selection if selection is None else selection
             schema = table.schema if schema is None else schema
@@ -843,6 +875,7 @@ class sql_cursor(object):
             writer = table.writer if writer is None else writer
             pk = table.pk if pk is None else pk
             doc = table.doc if doc is None else doc
+            defaults = table.defaults if defaults is None else defaults
             table = table.table
     
         self.table = table
@@ -850,6 +883,7 @@ class sql_cursor(object):
         self.db = db
         self.server = server
         self.engine = engine or get_engine(db = self.db, server = self.server)
+        self.session = session
         self.spec = spec
         self.selection = selection
         self.order = order
@@ -857,12 +891,125 @@ class sql_cursor(object):
         self.reader = reader
         self.writer = writer
         self.pk = pk
+        self.defaults = defaults
         self.doc = doc
     
     def copy(self):
         return type(self)(self)
 
+    def connect(self, dry_run = None, session_maker = None):
+        """
+        Creates a valid session and attach it to the cursor
+        
+        Parameters:
+        -----------
+        dry_run: bool
+            if set to True, when exiting the context, will rollback rather than commit
+        
+        session_maker: callable
+            a function that takes an engine and return a session-like object.
+            This allows user to implement their own logging etc. 
+            The resulting session must support 
+            * commit, rollback and execute
+            * context management (i.e. implement __enter__ and __exit__)            
+            
+        """
+        if session_maker is None:
+            session_maker = Session
+        if not valid_session(self.session):
+            self.session = session_maker(self.engine)
+        self.session.dry_run = dry_run
+        return self
+
     
+    def execute(self, statement, *args, transform = None, **kwargs):
+        """
+        executes a statement in two modes:
+            if a self.session exists, it assumes we are within a transaction and will simply execute using connection
+            if self.session is None, it assumes we basically want to lock and load... will execute and commit
+
+        Parameters
+        ----------
+        statement : sql statement
+        
+        Example: committed 
+        --------            
+        Example: a simple transactional logic
+        --------
+        with cursor:
+            cursor.execute(statement) ## not committed
+            cursor.execute(another_statement)
+
+        Example: a simple transactional logic for rolling back
+        --------
+        with cursor.connect(True):
+            cursor.execute(statement)
+            cursor.execute(another_statement)
+            
+
+        """        
+        if valid_session(self.session): ## we are within context
+            res = self.session.execute(statement, *args, **kwargs)
+            if transform:
+                res = transform(res)
+        else:
+            with self.engine.connect() as connection:
+                res = connection.execute(statement, *args, **kwargs)
+                if transform:
+                    res = transform(res)
+        return res
+    
+    def commit(self):
+        if valid_session(self.session):
+            if self.session.dry_run:
+                self.session.rollback()
+            else:
+                self.session.commit()
+        return self
+
+    def rollback(self):
+        if valid_session(self.session):
+            self.session.rollback()
+        return self
+        
+    def __enter__(self):
+        """
+        context manager entry point, creating an ORM session
+        
+        Example:
+        --------
+        with sql_table(..) as t:
+            t.insert()
+            t.delete()
+            
+        """
+        if not valid_session(self.session):
+            self.session = Session(self.engine)
+            self.session.dry_run = None
+        self.session.__enter__()
+        return self
+
+    def __exit__(self, type, value, traceback):
+        """
+        context manager exit point, removing the ORM session
+        
+        Example:
+        --------
+        with sql_table(..) as t:
+            t.insert()
+            t.delete()
+            
+        """
+        self.commit()
+        if valid_session(self.session):
+            self.session.__exit__(type, value, traceback)
+            self.session= None
+        return self
+    
+    def __del__(self):
+        if valid_session(self.session):
+            self.session.__del__()
+
     @property
     def _ids(self):
         """
@@ -908,7 +1055,6 @@ class sql_cursor(object):
         ids = self._ids
         return sorted([c.name for c in self.tbl.columns if c.nullable is False and c.name not in ids])
 
-
     @property
     def _columns(self):
         cols = self.columns
@@ -916,6 +1062,41 @@ class sql_cursor(object):
 
 
     def _kw(self, **kwargs):
+        """
+        converts kwargs dict into a sqlalchemy filtering expression
+        usually, we just use the column names
+        If the table is a document store and the key is not in the table columns, we assume the user wants to filter within the documnet
+        This is a little bit of a hack since we encode all documents but works quite well for simple queries
+        
+        Example: query within a document store
+        --------
+        >>> from pyg import * 
+        >>> table = partial(sql_table, db = 'test_db', schema = 'dbo', table = 'hello', nullable = 'a', pk = 'b', doc = True)
+        >>> c = db_cell(a = 'a', b = 'b', ccy = 'USD', asset = 'bond', db = table)()
+        >>> d = db_cell(a = 'b', b = 'c', ccy = 'EUR', asset = 'bond', db = table)()
+        >>> e = db_cell(a = 'a', b = 'd', ccy = 'EUR', asset = 'bond', db = table)()
+
+        >>> self = table()
+        >>> assert len(self.inc(ccy = ['USD','EUR'])) == 3
+        >>> assert len(self.inc(ccy = ['USD'])) == 1
+        >>> assert len(self.inc(ccy = 'EUR')) == 2
+
+
+        Example: query for a date within a doc store
+        --------------------------------------------
+        >>> from pyg import * 
+        >>> import datetime
+        >>> table = partial(sql_table, db = 'test_db', schema = 'dbo', table = 'dates', nullable = dict(a = datetime.datetime), pk = 'b', doc = True)
+        >>> c = db_cell(a = dt(2000), b = 'b', ccy = dt(2000), asset = 'bond', db = table)()
+        >>> d = db_cell(a = dt(2010), b = 'c', ccy = dt(2010), asset = 'bond', db = table)()
+        >>> e = db_cell(a = dt(2020), b = 'd', ccy = 3, asset = 'bond', db = table)()
+
+        >>> self = table()
+        >>> assert len(self.inc(ccy = dt(2000))) == 1
+        >>> assert len(self.inc(ccy = [dt(2000), dt(2010)])) == 2
+        >>> assert len(self.inc(ccy = 3)) == 1
+        
+        """
         columns = _get_columns(self._table)
         conditions = []
         for k, v in kwargs.items():
@@ -927,6 +1108,10 @@ class sql_cursor(object):
                 else:
                     col = col[0]
                 conditions.append(_pw_filter(col, v))
+            elif self.doc:
+                col = columns[self.doc][0]
+                vs = _like_within_doc(v, k)
+                conditions.append(_pw_filter(col, vs))
             else:
                 raise ValueError(f'column {k} not found')
         return sa.and_(*conditions)
@@ -938,6 +1123,7 @@ class sql_cursor(object):
         
         :Example:
         ---------
+        
         >>> expression = dict(a = 1, b = 2)
         >>> assert t._c(expression) == sa.and_(t.c.a == 1, t.c.b == 2)
         c = table.insert_many(dictable(a = ['a','aa','b', 'c'], b = ['bb','b1','c','d']))
@@ -1119,16 +1305,11 @@ class sql_cursor(object):
         return res
     
     
-    @property
-    def session(self):
-        return Session(self.engine)
-    
-    
     def __len__(self):
         statement = select(func.count()).select_from(self._table)
         if self.spec is not None:
             statement = statement.where(self.spec)
-        return list(self.engine.connect().execute(statement))[0][0]
+        return self.execute(statement, transform = list)[0][0]
     
     count = __len__
 
@@ -1199,7 +1380,9 @@ class sql_cursor(object):
             Suppose you have a document with EXTRA keys. Rather than filter the document, set ignore_bad_keys = True and we will drop irrelevant keys for you
 
         """
-        doc = _relabel(res = doc, selection = self.columns, strict = False) ## we want to replace what is possible since 
+        if self.defaults:
+            doc.update({k : v for k,v in self.defaults.items() if k not in doc})
+        doc = _relabel(res = doc, selection = self.columns, strict = False) ## we want to replace what is possible
         edoc = self._dock(doc) if write else doc
         columns = self.columns
         if not ignore_bad_keys:
@@ -1213,10 +1396,9 @@ class sql_cursor(object):
             res_no_ids = type(res)({k : v for k, v in res.items() if k not in ids}) if ids else res
             tbl = self.inc().inc(**doc_id)
             read = tbl.sort(ids)._read_statement() ## row format
-            docs = tbl._rows_to_docs(reader = False, load = False, **read) ## do not transform the document, keep in raw format
+            docs = tbl._rows_to_docs(reader = False, load = False, **read) ## do not transform the document, keep in raw format?
             if len(docs) == 0:
-                with self.engine.connect() as conn: 
-                    conn.execute(self.table.insert(),[res_no_ids])
+                self.execute(self.table.insert(),[res_no_ids])
                 if ids:    
                     latest = tbl[0]
                     doc.update(latest[ids])
@@ -1237,8 +1419,7 @@ class sql_cursor(object):
                 self.inc(self._id(latest)).update(**(res_no_ids))
                 doc.update(latest[ids])
         else:
-            with self.engine.connect() as conn:
-                conn.execute(self.table.insert(), [res])
+            self.execute(self.table.insert(), [res])
         return doc
     
     
@@ -1345,13 +1526,12 @@ class sql_cursor(object):
             stop = stop if stop is None else stop - start
         if stop is not None:
             statement = statement.limit(1+stop)
-        executed = self.engine.connect().execute(statement)
-        columns = list(executed.keys())
-        data = list(executed)
+        data, columns = self.execute(statement, transform = _data_and_columns)
         if start is not None or stop is not None or step is not None:
             data = data[slice(start, stop, step)]
         return dict(data = data, columns = columns)
     
+    
     def df(self, decimal2float = True):
         """
         This is a more optimized, faster version for reading the table. 
@@ -1389,12 +1569,11 @@ class sql_cursor(object):
 
         """
         statement = self.statement()
-        res = self.engine.connect().execute(statement)
-        columns = list(res.keys())
         if _pd_is_old:
-            res = pd.DataFrame(list(res), columns = columns)
+            data, columns = self.execute(statement, transform = _data_and_columns)
+            res = pd.DataFrame(data, columns = columns)
         else:
-            res = pd.DataFrame(res)
+            res = self.execute(statement, transform = pd.DataFrame)
         if decimal2float:
             for col in res.columns:
                 t = self.table.columns[col].type
@@ -1424,14 +1603,10 @@ class sql_cursor(object):
         value = len(self) + item if item < 0 else item
         statement = self.statement()
         if self.order is None:
-            executed = self.engine.connect().execute(statement.limit(value+1))
-            columns = list(executed.keys())
-            data = list(executed)
+            data, columns = self.execute(statement.limit(value+1), transform = _data_and_columns)
             row = data[value]
         else:
-            executed = self.engine.connect().execute(statement.offset(value).limit(1))
-            columns = list(executed.keys())
-            data = list(executed)
+            data, columns = self.execute(statement.limit(value+1), transform = _data_and_columns)
             row = data[0]
         doc = self._rows_to_docs(data = row, reader = reader, columns = columns)
         rtn = self._undock(doc)
@@ -1494,6 +1669,8 @@ class sql_cursor(object):
         """
         Similar to insert, except will throw an error if upsert = False and an existing document is not there
         """
+        if self.defaults:
+            doc.update({k : v for k,v in self.defaults.items() if k not in doc})
         existing = self.inc().inc(**self._id(doc))
         n = len(existing)
         if n == 0:
@@ -1527,6 +1704,8 @@ class sql_cursor(object):
         docs : list of dicts, dictable, pd.DataFrame
         """
         rs = dictable(docs)
+        if self.defaults:
+            rs = rs(**{k : v for k,v in self.defaults.items() if k not in rs.keys()})
         rs = _relabel(rs, self.columns, strict = False)
         if len(rs) > 0:
             if self._pk and not self._is_deleted():
@@ -1538,8 +1717,7 @@ class sql_cursor(object):
                     rows = [self._write_doc(row, columns = columns) for row in rows]
                 else:
                     rows = list(rs)
-                with self.engine.connect() as conn:
-                    conn.execute(self.table.insert(), rows)
+                self.execute(self.table.insert(), rows)
         return self
 
     def __iter__(self):
@@ -1573,6 +1751,8 @@ class sql_cursor(object):
         
         """
         rs = dictable(data = data, columns = columns, **kwargs) ## this allows us to insert multiple rows easily as well as pd.DataFrame
+        if self.defaults:
+            rs = rs(**{k : v for k,v in self.defaults.items() if k not in rs.keys()})
         if len(rs)>1:
             pk = self._pk
             if pk:
@@ -1679,8 +1859,7 @@ class sql_cursor(object):
         if self.spec is not None:
             statement = statement.where(self.spec)
         statement = statement.values(kwargs)
-        with self.engine.connect() as conn:
-            conn.execute(statement)
+        self.execute(statement)
         return self
     
     set = update
@@ -1694,8 +1873,7 @@ class sql_cursor(object):
         statement = self.table.delete()
         if self.spec is not None:
             statement = statement.where(self.spec)
-        with self.engine.connect() as conn:
-            conn.execute(statement)
+        self.execute(statement)
         return self
 
     def delete(self, **kwargs):
@@ -1705,6 +1883,7 @@ class sql_cursor(object):
             if self._pk and not self._is_deleted(): ## we first copy the existing data out to deleted db
                 read = self._read_statement() 
                 docs = self._rows_to_docs(reader = False, load = False, **read)
+                #docs = self._rows_to_docs(load = True, **read)
                 deleted = datetime.datetime.now()
                 for doc in docs:
                     doc[_deleted] = deleted
@@ -1780,16 +1959,22 @@ class sql_cursor(object):
     def deleted(self):
         if self._is_deleted() or len(self._pk) == 0:
             return self
-        else:        
+        else:
             schema = _archived + (self.schema or '')
             # logger.info('archived schema: %s'%schema)
+            writer = self.writer
+            if is_str(writer) and writer.endswith('.sql'):
+                params = writer.split('/')
+                params[2] = schema
+                writer = '/'.join(params)
+                writer = writer.replace('.sql','/%deleted.sql')
             res = sql_table(table = self.table, 
                             db = self.db, 
                             non_null = dict(deleted = datetime.datetime), 
                             server = self.server, 
-                            pk = self.pk, 
+                            pk = self._pk + [_deleted], 
                             doc = self.doc, 
-                            writer = self.writer, 
+                            writer = writer, 
                             reader = self.reader, 
                             schema = schema)
             #res.spec = self.spec THIS NEEDS TO BE IMPLEMENTED PROPERLY
